<?php

declare(strict_types=1);

namespace Plugin\lfs_shopvote\classes;

use Carbon\Carbon;
use JTL\Catalog\Product\Artikel;
use JTL\DB\DbInterface;
use JTL\DB\ReturnType;
use JTL\Plugin\PluginInterface;
use JTL\Services\JTL\Validation\Rules\DateTime;
use JTL\Shop;

class LfsShopvote
{
    /**
     * @var PluginInterface
     */
    private $plugin;

    /**
     * @var DbInterface
     */
    private $DB;

    /**
     * @var string
     */
    private $api_key;

    /**
     * @var string
     */
    private $api_secret;

    /**
     * @var string
     */
    private $api_url = 'https://api.shopvote.de/';

    /**
     * @var string
     */
    private $auth_token = "";

    /**
     * @var string
     */
    private $auth_token_validity;

    /**
     * @var string|null
     */
    private $shopid = null;

    /**
     * @var string
     */
    private $useragent = 'App.T6Z';

    /**
     * @var string[]
     */
    private $valid_settings = [
        'sv_show_productreview' => 'pluginsetting',
        'sv_api_key' => 'pluginsetting',
        'sv_api_secret' => 'pluginsetting',
        'sv_graphic_code' => 'extendedsetting',
        'sv_respect_jtlconsentmanager' => 'pluginsetting'
    ];

    /**
     * @var string
     */
    private $graphicsCode;


    public function __construct(PluginInterface $plugin, DbInterface $db)
    {
        $this->setPlugin($plugin);
        $this->setDB($db);
        $this->setApiKey($this->getPlugin()->getConfig()->getValue('sv_api_key'));
        $this->setApiSecret($this->getPlugin()->getConfig()->getValue('sv_api_secret'));

        if (!isset($_SESSION['lfsShopVoteAuthToken'])) {
            $_SESSION['lfsShopVoteAuthToken'] = "";
            $_SESSION['lfsShopVoteAuthTokenValidity'] = 0;
        }

        if (isset($_SESSION['lfsShopVoteAuthToken'])) {
            $this->setAuthToken($_SESSION['lfsShopVoteAuthToken']);
        }

        if (isset($_SESSION['lfsShopVoteShopID'])) {
            $this->setShopid($_SESSION['lfsShopVoteShopID']);
        }
    }

    /**
     * @return PluginInterface
     */
    public function getPlugin(): PluginInterface
    {
        return $this->plugin;
    }

    /**
     * @param PluginInterface $plugin
     * @return LfsShopvote
     */
    public function setPlugin(PluginInterface $plugin): LfsShopvote
    {
        $this->plugin = $plugin;
        return $this;
    }

    /**
     * @return string
     */
    public function getApiKey(): string
    {
        return $this->api_key;
    }

    /**
     * @param string $api_key
     * @return LfsShopvote
     */
    public function setApiKey(string $api_key): LfsShopvote
    {
        $this->api_key = $api_key;
        return $this;
    }

    /**
     * @return string
     */
    public function getApiSecret(): string
    {
        return $this->api_secret;
    }

    /**
     * @param string $api_secret
     * @return LfsShopvote
     */
    public function setApiSecret(string $api_secret): LfsShopvote
    {
        $this->api_secret = $api_secret;
        return $this;
    }

    /**
     * @return string
     */
    public function getApiUrl(): string
    {
        return $this->api_url;
    }

    /**
     * @param string $api_url
     * @return LfsShopvote
     */
    public function setApiUrl(string $api_url): LfsShopvote
    {
        $this->api_url = $api_url;
        return $this;
    }

    /**
     * @return string
     */
    public function getAuthToken(): string
    {
        return $this->auth_token;
    }

    /**
     * @param string $auth_token
     * @return LfsShopvote
     */
    public function setAuthToken(string $auth_token): LfsShopvote
    {
        $_SESSION['lfsShopVoteAuthToken'] = $auth_token;
        $this->auth_token = $auth_token;
        return $this;
    }

    /**
     * @return int
     */
    public function getAuthTokenValidity(): int
    {
        return $this->auth_token_validity = 0;
    }

    /**
     * @param int $auth_token_validity
     * @return LfsShopvote
     */
    public function setAuthTokenValidity(int $auth_token_validity): LfsShopvote
    {
        $this->auth_token_validity = $auth_token_validity;
        return $this;
    }

    /**
     * @return string|null
     */
    public function getShopid(): ?string
    {
        return $this->shopid;
    }

    /**
     * @param string $shopid
     * @return LfsShopvote
     */
    public function setShopid(string $shopid): LfsShopvote
    {
        $_SESSION['lfsShopVoteShopID'] = $shopid;
        $this->shopid = $shopid;
        return $this;
    }

    /**
     * @return string
     */
    public function getUseragent(): string
    {
        return $this->useragent;
    }

    /**
     * @param string $useragent
     * @return LfsShopvote
     */
    public function setUseragent(string $useragent): LfsShopvote
    {
        $this->useragent = $useragent;
        return $this;
    }

    /**
     * @return DbInterface
     */
    public function getDB(): DbInterface
    {
        return $this->DB;
    }

    /**
     * @param DbInterface $DB
     * @return LfsShopvote
     */
    public function setDB(DbInterface $DB): LfsShopvote
    {
        $this->DB = $DB;
        return $this;
    }

    /**
     * @return string[]
     */
    public function getValidSettings(): array
    {
        return $this->valid_settings;
    }

    /**
     * @param string[] $valid_settings
     * @return LfsShopvote
     */
    public function setValidSettings(array $valid_settings): LfsShopvote
    {
        $this->valid_settings = $valid_settings;
        return $this;
    }

    /**
     * @return string
     */
    public function getGraphicsCode(): string
    {
        return $this->graphicsCode;
    }

    /**
     * @param string $graphicsCode
     * @return LfsShopvote
     */
    public function setGraphicsCode(string $graphicsCode): LfsShopvote
    {
        $this->graphicsCode = $graphicsCode;
        return $this;
    }

    public function renderConfigTab(): string
    {
        Shop::Smarty()->assign('plugin', $this->getPlugin());

        foreach ($this->getValidSettings() as $setting => $type)
        {
            if ($type === 'pluginsetting')
            {
                $currentValue = $this->getDB()->selectSingleRow('tplugineinstellungen', 'cName', $setting);
                Shop::Smarty()->assign($setting, $currentValue->cWert);
            }
            else
            {
                $currentValue = $this->getDB()->selectSingleRow('xplugin_lfs_shopvote_config', 'cName', $setting);
                Shop::Smarty()->assign($setting, $currentValue->cWert);
            }
        }

        return Shop::Smarty()->fetch($this->getPlugin()->getPaths()->getAdminPath() . '/template/lfs_shopvote_config.tpl');
    }

    public function saveSettings($post): bool
    {
        $bSuccess = true;

        foreach ($this->getValidSettings() as $setting => $type)
        {
            if ($type === 'pluginsetting')
            {
                $this->getDB()->executeQueryPrepared('UPDATE tplugineinstellungen SET cWert = :value 
                                                                        WHERE 
                                                                              cName = :settingname
                                                                        AND 
                                                                            kPlugin = :kPlugin',
                    [
                        'settingname' => $setting,
                        'value' => $post[$setting],
                        'kPlugin' => $this->getPlugin()->getID()
                    ], ReturnType::AFFECTED_ROWS);
            }
            else
            {
                $this->getDB()->executeQueryPrepared('UPDATE xplugin_lfs_shopvote_config SET cWert = :value 
                                                                        WHERE 
                                                                              cName = :settingname',
                    [
                        'settingname' => $setting,
                        'value' => $post[$setting]
                    ], ReturnType::AFFECTED_ROWS);
            }
        }

        return $bSuccess;
    }

    public function returnGraphicsCode():string
    {
        $graphicsCodeObject = $this->getDB()->selectSingleRow('xplugin_lfs_shopvote_config', 'cName', 'sv_graphic_code');
        return $graphicsCodeObject->cWert;
    }

    public function authorize():void
    {
        if (
            $_SESSION['lfsShopVoteAuthTokenValidity'] === 0 ||
            $_SESSION['lfsShopVoteAuthTokenValidity'] < time()-60 ||
            $this->getAuthToken() === ""
        )
        {
            $headers[] = 'Apikey: ' . $this->getApiKey();
            $headers[] = 'Apisecret: '. $this->getApiSecret();
            $headers[] = 'User-Agent' . $this->getUseragent();

            $token = $this->callShopvoteAPI($this->getApiUrl().'auth', $this->getUseragent(), $headers);

            if ($this->isJson($token)) {
                $this->decodeJwt($token);
                $this->setAuthToken(json_decode($token)->Token);
            }
            else
            {
                $logservice = Shop::Container()->getLogService();
                $logservice->info($this->getPlugin()->getPluginID() . ": " . $e->getMessage());
            }
        }
    }

    public function calculateCombinedAverageReviews($article, $review_data): object
    {
        $oReturn = new \stdClass();
        $oReturn->average_rating = 0;
        $tmpTotalRating = 0;

        $oReturn->review_count = count($article->Bewertungen->oBewertung_arr)+$review_data->rating_count;

        if (\is_array($article->Bewertungen->oBewertung_arr) && count($article->Bewertungen->oBewertung_arr) > 0) {
            foreach ($article->Bewertungen->oBewertung_arr as $review)
            {
                $tmpTotalRating += $review->nSterne;
                $oReturn->rating_count++;
            }
        }

        if (\is_array($review_data->reviews) && count($review_data->reviews) > 0) {
            foreach ($review_data->reviews as $review)
            {
                $tmpTotalRating += $review->rating_value;
                $oReturn->rating_count++;
            }
        }

        if ($oReturn->review_count>0)
        {
            $oReturn->average_rating = $tmpTotalRating/$oReturn->review_count;
        }

        return $oReturn;
    }

    public function calculateAverageReviews($review_id): void
    {
        $oReturn = new \stdClass();
        $oReturn->average_rating = 0;
        $oReturn->rating_count = 0;
        $tmpTotalRating = 0;

        $reviews = $this
            ->getDB()
            ->executeQueryPrepared(
                'SELECT * FROM xplugin_lfs_shopvote_article_review_items WHERE review_id = :review_id',
                    [
                        'review_id' => $review_id
                    ],
            ReturnType::ARRAY_OF_OBJECTS);

        if (\is_array($reviews) && count($reviews) > 0) {
            foreach ($reviews as $review)
            {
                $tmpTotalRating += $review->rating_value;
                $oReturn->rating_count++;
            }
        }

        $oReturn->review_count = count($reviews);

        if ($oReturn->review_count>0)
        {
            $oReturn->average_rating = \round($tmpTotalRating/$oReturn->review_count,2);
        }

        $this->getDB()->executeQueryPrepared('UPDATE xplugin_lfs_shopvote_article_reviews 
                                                        SET  
                                                            rating_count = :rating_count, 
                                                            rating_value = :rating_value,  
                                                            updated_at = :updated_at
                                                        WHERE 
                                                            id = :review_id',
            [
                'review_id' => $review_id,
                'rating_count' => $oReturn->review_count,
                'rating_value' => $oReturn->average_rating,
                'updated_at' => Carbon::now()->addMinutes(5)->format('Y-m-d H:i:s')
            ], ReturnType::AFFECTED_ROWS);
    }

    private function calculateAverageRatingsByRatingitems(&$productReviewData)
    {
        $tmpTotalRating = 0;
        $tmpRatingCount = 0;
        $tmpAverageRating = 0;

        if (\is_array($productReviewData->reviews) && count($productReviewData->reviews) > 0) {
            foreach ($productReviewData->reviews as $review)
            {
                $tmpRatingCount++;
                $tmpTotalRating += $review->rating_value;
            }
        }

        $tmpReviewCount = count($productReviewData->reviews);

        if ($tmpReviewCount>0)
        {
            $tmpAverageRating = \round($tmpTotalRating/$tmpReviewCount,2);
        }

        $productReviewData->rating_count = $tmpRatingCount;
        $productReviewData->rating_value = $tmpAverageRating;
    }

    public function getProductReviews(Artikel $article):object
    {
        $productReviewData = new \stdClass();
        $productReviewData->reviews = [];
        $productReviewData->rating_count = 0;
        $productReviewData->rating_value = 0;

        $ids = [];

        if ($this->getApiKey() !== '' && $this->getApiSecret() !== '')
        {
            if ((int)$article->kVaterArtikel === 0)
            {
                $article_ids = $this
                    ->getDB()
                    ->executeQueryPrepared(
                        'SELECT kArtikel FROM tartikel WHERE kVaterartikel = :kVaterartikel AND kArtikel IN (
                            SELECT article_id
                            FROM xplugin_lfs_shopvote_article_reviews
                            WHERE updated_at < DATE_SUB(NOW(), INTERVAL 6 HOUR)
                        )',
                        [
                            'kVaterartikel' => $article->kArtikel
                        ],
                        ReturnType::ARRAY_OF_OBJECTS
                    );

                if (\is_array($article_ids) && \count($article_ids)>0)
                {
                    foreach ($article_ids as $article_id) {
                        $ids[] = $article_id->kArtikel;
                    }
                }
                else {
                    $ids[] = $article->kArtikel;
                }
            }
            else {
                $ids[] = $article->kArtikel;
            }

            if (\is_array($ids) && \count($ids)>0)
            {
                foreach ($ids as $id) {
                    $tmpArticle = new Artikel();
                    $tmpArticle->fuelleArtikel((int)$id);

                    $bReloadReviewInfo = false;

                    $reviewInfo = $this->getDB()->selectSingleRow('xplugin_lfs_shopvote_article_reviews', 'article_id', $id);
                    if ($reviewInfo === null)
                    {
                        // Call ShopVote-API and get data
                        $reviewData = $this->syncProductReview($tmpArticle);

                        if ($this->isJson($reviewData))
                        {
                            $bReloadReviewInfo = true;
                            $this->saveReviewInfo($tmpArticle, json_decode($reviewData));
                        } else {
                            $reviewInfo = new \stdClass();
                        }
                    }
                    else
                    {
                        $reviewData = $this->syncProductReview($tmpArticle);

                        if ($this->isJson($reviewData))
                        {
                            $bReloadReviewInfo = true;
                            $this->saveReviewInfo($article, json_decode($reviewData));
                        } else {
                            $reviewInfo = new \stdClass();
                        }
                    }

                    if ($bReloadReviewInfo)
                    {
                        $reviewInfo = $this->getDB()->selectSingleRow('xplugin_lfs_shopvote_article_reviews', 'article_id', $id);
                    }

                    if (isset($reviewInfo->id))
                    {
                        if (\count($ids) === 1)
                        {
                            $productReviewData = $reviewInfo;
                        }

                        $reviewItems = $this
                            ->getDB()
                            ->executeQueryPrepared('SELECT * FROM xplugin_lfs_shopvote_article_review_items WHERE review_id = :review_id ORDER BY created_at DESC',
                                [
                                    'review_id' => $reviewInfo->id
                                ],
                                ReturnType::ARRAY_OF_OBJECTS);

                        if ($reviewItems !== null)
                        {
                            foreach ($reviewItems as $reviewItem) {
                                $productReviewData->reviews[] = $reviewItem;
                            }
                        }
                    }
                }
            }
        }

        if (\count($ids)>1)
        {
            $this->calculateAverageRatingsByRatingitems($productReviewData);
        }

        if (\is_array($productReviewData->reviews) && \count($productReviewData->reviews)>20)
        {
            // Limit to 20 reviewitems
            $productReviewData->reviews = \array_slice($productReviewData->reviews, 0, 20);
        }

        return $productReviewData;
    }

    public function getIdByArtNr($cArtNr)
    {
        $tmpArticle = $this->getDB()->executeQueryPrepared(
            'SELECT kArtikel FROM tartikel WHERE cArtNr = :cArtNr',
        [
            'cArtNr' => $cArtNr
        ],
        ReturnType::SINGLE_OBJECT);

        if (\is_object($tmpArticle)) {
            return (int)$tmpArticle->kArtikel;
        }

        return 0;
    }

    private function syncProductReview($article):string
    {
        try {
            $this->authorize();

            if ($this->getShopid() !== null)
            {
                $headers[] = 'Token: Bearer ' . $this->getAuthToken();
                $headers[] = 'User-Agent' . $this->getUseragent() . "." . $this->getShopid();

                $params['sku'] = \urlencode($article->cArtNr);
                $params['sd'] = 'false';

                $apiResult = $this->callShopvoteAPI($this->getApiUrl().'product-reviews/v2/reviews?', $this->getUseragent() . '.' . $this->getShopid(), $headers, $params);

                if (empty($apiResult)) {
                    return '';
                } else {
                    return $apiResult;
                }
            }
            else
            {
                $logservice = Shop::Container()->getLogService();
                $logservice->debug($this->getPlugin()->getPluginID() . ': Connection to Shopvote-API could not be established');
                return '';
            }
        } catch (\Exception $e) {
            $logservice = Shop::Container()->getLogService();
            $logservice->debug($this->getPlugin()->getPluginID() . ': ' . $e->getMessage());
            return '';
        }
    }

    private function saveReviewInfo($article, $reviewData):void
    {
        $review = $this->getDB()->selectSingleRow('xplugin_lfs_shopvote_article_reviews', 'article_id', $article->kArtikel);

        if ($review === null)
        {
            $productReview_id = $this->getDB()->executeQueryPrepared('INSERT INTO xplugin_lfs_shopvote_article_reviews 
                                                        (article_id, reviewpage, rating_count, rating_value, created_at, updated_at) 
                                                        VALUES 
                                                        (:article_id, :reviewpage, :rating_count, :rating_value, :created_at, :updated_at)',
                [
                    'article_id' => $article->kArtikel,
                    'reviewpage' => $reviewData->reviewpage,
                    'rating_count' => $reviewData->rating_count,
                    'rating_value' => $reviewData->rating_value,
                    'created_at' => Carbon::now()->format('Y-m-d H:i:s'),
                    'updated_at' => Carbon::now()->format('Y-m-d H:i:s')
                ], ReturnType::LAST_INSERTED_ID);
        }
        else
        {
            $productReview_id = $review->id;
            $this->getDB()->executeQueryPrepared('UPDATE xplugin_lfs_shopvote_article_reviews 
                                                        SET  
                                                            reviewpage = :reviewpage, 
                                                            rating_count = :rating_count, 
                                                            rating_value = :rating_value, 
                                                            updated_at = :updated_at
                                                        WHERE 
                                                            article_id = :article_id',
                [
                    'article_id' => $article->kArtikel,
                    'reviewpage' => $reviewData->reviewpage,
                    'rating_count' => $reviewData->rating_count,
                    'rating_value' => $reviewData->rating_value,
                    'updated_at' => Carbon::now()->format('Y-m-d H:i:s')
                ], ReturnType::AFFECTED_ROWS);
        }

        if ($reviewData->rating_count > 0) {
            $this->saveProductReviewItems($productReview_id, $reviewData->reviews);
            $this->calculateAverageReviews($productReview_id);
        }
    }

    private function saveProductReviewItems($productReview_id, $items)
    {
        $ids = [];
        $existingReviewsTmp = $this->getDB()->executeQuery('SELECT id FROM xplugin_lfs_shopvote_article_review_items 
                                                                WHERE
                                                                  review_id = ' . $productReview_id, ReturnType::ARRAY_OF_OBJECTS);

        if (\is_array($existingReviewsTmp) && count($existingReviewsTmp)>0)
        {
            foreach ($existingReviewsTmp as $reviewItem)
            {
                $ids[] = $reviewItem->id;
            }
        }

        foreach ($items as $item)
        {
            if (!\in_array($item->reviewUID, $ids))
            {
                $this->getDB()->executeQueryPrepared('INSERT INTO xplugin_lfs_shopvote_article_review_items 
                                                        (id, review_id, author, rating_value, rating_text, created_at) 
                                                        VALUES 
                                                        (:id, :review_id, :author, :rating_value, :rating_text, :created_at)',
                    [
                        'id' => $item->reviewUID,
                        'review_id' => $productReview_id,
                        'author' => $item->author,
                        'rating_value' => $item->rating_value,
                        'rating_text' => $item->text,
                        'created_at' => $item->created
                    ], ReturnType::AFFECTED_ROWS);
            }
        }
    }

    private function callShopvoteAPI($url, $useragent, $headers, $params = null)
    {
        if ($params !== null)
        {
            $count = 0;
            foreach ($params as $k => $v) {
                if ($count > 0)
                {
                    $url .= '&';
                }
                $url .= $k . '=' . $v;
                $count++;
            }
        }

        $ch = \curl_init();
        \curl_setopt($ch, CURLOPT_URL, $url);
        \curl_setopt($ch, CURLOPT_TIMEOUT_MS, 1500);
        \curl_setopt($ch, CURLOPT_CONNECTTIMEOUT_MS, 1000);
        \curl_setopt($ch, CURLOPT_RETURNTRANSFER, 1);
        \curl_setopt($ch, CURLOPT_USERAGENT, $useragent);
        \curl_setopt($ch, CURLINFO_HEADER_OUT, true);

        if (\is_array($headers)) {
            \curl_setopt($ch, CURLOPT_HTTPHEADER, $headers);
        }

        $result = \curl_exec($ch);

        $response_code = \curl_getinfo($ch, CURLINFO_RESPONSE_CODE);

        if ((int)$response_code !== 200)
        {
            $logservice = Shop::Container()->getLogService();
            $logservice->debug($this->getPlugin()->getPluginID() . ': Got HTTP-Code: '. (int)$response_code. ' from ShopVote-API');
            return null;
        }

        if ($this->isJson($result)) {
            $resultObj = json_decode($result);

            if ((isset($resultObj->Code)) && ((string)$resultObj->Code !== '200'))  {
                $logservice = Shop::Container()->getLogService();
                $logservice->info($this->getPlugin()->getPluginID() . ': Code: '.$resultObj->Code.' received from ShopVote-API - Info: '.$resultObj->Message);

                return null;
            }
            else
            {
                return $result;
            }

        }

        $logservice = Shop::Container()->getLogService();
        $logservice->info($this->getPlugin()->getPluginID() . ': No valid data received from ShopVote-API');

        return null;
    }

    private function isJson($string): bool
    {
        return \is_string($string) && \is_array(json_decode($string, true)) && (\json_last_error() === JSON_ERROR_NONE);
    }

    private function decodeJwt($token): void
    {
        $decoded = json_decode(
            \base64_decode(
                str_replace('_', '/', str_replace('-', '+', explode('.', $token)[1]))
            )
        );

        $this->setShopid($decoded->shopid);
        $_SESSION['lfsShopVoteAuthTokenValidity'] = $decoded->exp;
    }

    /**
     * @param int $days
     * @return void
     */
    public function getArticlesToUpdate(int $days = 2)
    {
        try {
            $this->authorize();

            if ($this->getShopid() !== null)
            {
                $headers[] = 'Token: Bearer ' . $this->getAuthToken();
                $headers[] = 'User-Agent' . $this->getUseragent() . "." . $this->getShopid();

                $params['sd'] = 'false';
                $params['days'] = $days;

                $apiResult = $this->callShopvoteAPI($this->getApiUrl().'product-reviews/v2/reviews?', $this->getUseragent() . '.' . $this->getShopid(), $headers, $params);

                if (!empty($apiResult)) {
                    $tmpData = json_decode($apiResult);
                    if (isset($tmpData->reviews))
                    {
                        $reviewData = collect(json_decode($apiResult)->reviews);
                        $reviewData = $reviewData->unique(['sku'])->pluck('sku')->toArray();

                        if (is_array($reviewData) && count($reviewData)>0)
                        {
                            foreach ($reviewData as $reviewItem) {
                                $sql = "INSERT INTO xplugin_lfs_shopvote_reviewupdate_queue (kartikel, sku, added, updated)
                                            SELECT kArtikel, '" . $reviewItem . "', '" . Carbon::now() ."', '" . Carbon::now() ."' FROM tartikel 
                                                WHERE cArtNr = '" . $reviewItem . "'
                                        ON DUPLICATE KEY UPDATE
                                            updated = '" . Carbon::now() ."'";

                                $this->getDB()->executeExQuery($sql);
                            }
                        }
                    }
                    else
                    {
                        dump("No Data");
                    }
                }
            }
            else
            {
                $logservice = Shop::Container()->getLogService();
                $logservice->debug($this->getPlugin()->getPluginID() . ': Connection to Shopvote-API could not be established');
            }
        } catch (\Exception $e) {
            $logservice = Shop::Container()->getLogService();
            $logservice->debug($this->getPlugin()->getPluginID() . ': ' . $e->getMessage());
        }
    }
}